"""
Utilities for testing
"""
import sys
import numpy as np
from tables_io.lazy_modules import LazyModule, tables, apTable, apDiffUtils, fits, h5py, pd, pq, jnp


def check_deps(deps=None):
    missing = False
    if deps is None:  # pragma: no cover
        deps = [tables, apTable, apDiffUtils, fits, h5py, pd, pq, jnp]
    for mod in deps:
        if not isinstance(mod, LazyModule):
            sys.stderr.write(f"{mod} is not LazyModule: {type(mod)}")
            missing = True
            continue
        try:
            _ = dir(mod)
        except Exception as err:  # pylint: disable=broad-exception-caught
            sys.stderr.write(f"Missing {mod.name} because {err}")
            missing = True
    return not missing


def compare_tables(t1, t2, columns=None):
    """Compare all the tables in two `astropy.table.Table`)

    Parameters
    ----------
    t1 : `astropy.table.Table`
       One table
    t2 : `astropy.table.Table`
       Another tables

    Returns
    -------
    identical : `bool`
        True if the tables are identical, False otherwise

    Notes
    -----
    For now this explicitly flattens each of the columns, to avoid issues with shape
    """
    if columns:
        t1.keep_columns(columns)
        t2.keep_columns(columns)
    if sorted(t1.colnames) != sorted(t2.colnames):  # pragma: no cover
        return False
    for cname in t1.colnames:
        c1 = t1[cname]
        c2 = t2[cname]
        if not np.allclose(np.array(c1).flat, np.array(c2).flat):  # pragma: no cover
            return False
    return True


def compare_table_dicts(d1, d2, strict=False, columns=None):
    """Compare all the tables in two `OrderedDict`, (`str`, `astropy.table.Table`)

    Parameters
    ----------
    d1 : `OrderedDict`, (`str`, `astropy.table.Table`)
       One dictionary of tables
    d2 : `OrderedDict`, (`str`, `astropy.table.Table`)
       Another dictionary of tables

    Returns
    -------
    identical : `bool`
        True if all the tables are identical, False otherwise
    """
    identical = True
    for k, v in d1.items():
        try:
            vv = d2[k]
        except KeyError:  # pragma: no cover
            vv = d2[k.upper()]
        if strict:  # pragma: no cover
            identical &= apDiffUtils.report_diff_values(v, vv)
        else:  # pragma: no cover
            identical &= compare_tables(v, vv, columns=columns)
    return identical


def make_test_data():
    """Make and return some test data"""
    nrow = 1000
    vect_size = 20
    mat_size = 5
    scalar = np.random.uniform(size=nrow)
    vect = np.random.uniform(size=nrow * vect_size).reshape(nrow, vect_size)
    matrix = np.random.uniform(size=nrow * mat_size * mat_size).reshape(nrow, mat_size, mat_size)
    data = dict(scalar=scalar, vect=vect, matrix=matrix)
    table = apTable.Table(data)
    table.meta["a"] = 1
    table.meta["b"] = None
    table.meta["c"] = [3, 4, 5]
    small_table = apTable.Table(dict(a=np.ones(21), b=np.zeros(21)))
    small_table.meta["small"] = True
    out_tables = dict(data=table, md=small_table)
    return out_tables



""" Unit tests for the fileIO module """

import sys
import unittest

import numpy as np
import pytest

from tables_io import arrayUtils, types
from tables_io.lazy_modules import apTable, lazyImport, pd
from tables_io.testUtils import check_deps


def test_check_deps():
    bad_module = lazyImport('this_does_not_exist')
    assert not check_deps([bad_module])


def test_array_length():
    """Test the pandas reading"""
    assert arrayUtils.arrayLength(4) == 0
    assert arrayUtils.arrayLength(np.ones(5)) == 5
    assert arrayUtils.arrayLength(np.ones((5, 5, 5))) == 5
    assert arrayUtils.arrayLength([3, 4, 4]) == 3


def test_force_to_pandasable():
    """Test the force_to_pandasable function"""
    rv = np.random.uniform(size=100)
    assert arrayUtils.forceToPandables(4) == 4
    assert np.allclose(arrayUtils.forceToPandables(rv), rv)
    try:
        arrayUtils.forceToPandables(rv, 95)
    except ValueError:
        pass
    else:
        raise ValueError("Failed to catch array length mismatch in arrayUtils.forceToPandables")
    rv2d = rv.reshape(10, 10)
    rv2d_check = np.vstack(arrayUtils.forceToPandables(rv2d))
    assert np.allclose(rv2d, rv2d_check)

    rv3d = rv.reshape(10, 2, 5)
    rv3d_check = np.vstack(arrayUtils.forceToPandables(rv3d)).reshape(rv3d.shape)
    assert np.allclose(rv3d, rv3d_check)


def test_slice_dict():
    """Test the slice_dict method"""
    test_data = dict(
        scalar=np.random.uniform(size=10),
        vector=np.random.uniform(size=100).reshape(10, 10),
        mat=np.random.uniform(size=1000).reshape(10, 10, 10),
    )
    sliced = arrayUtils.sliceDict(test_data, 1)
    assert np.allclose(sliced["scalar"], test_data["scalar"][1])
    assert np.allclose(sliced["vector"], test_data["vector"][1])
    assert np.allclose(sliced["mat"], test_data["mat"][1])

    mask = np.zeros((10), bool)
    mask[1] = True
    mask[4] = True
    sliced = arrayUtils.sliceDict(test_data, mask)
    assert np.allclose(sliced["scalar"], test_data["scalar"][mask])
    assert np.allclose(sliced["vector"], test_data["vector"][mask])
    assert np.allclose(sliced["mat"], test_data["mat"][mask])


def test_print_dict_shape():
    """Test the print_dict_shape method"""
    test_data = dict(
        scalar=np.random.uniform(size=10),
        vector=np.random.uniform(size=100).reshape(10, 10),
        mat=np.random.uniform(size=1000).reshape(10, 10, 10),
    )
    arrayUtils.printDictShape(test_data)


def test_concatenateDicts():
    """Test the print_dict_shape method"""
    test_data = dict(
        scalar=np.random.uniform(size=10),
        vector=np.random.uniform(size=100).reshape(10, 10),
        mat=np.random.uniform(size=1000).reshape(10, 10, 10),
    )
    od = arrayUtils.concatenateDicts([test_data, test_data])
    assert np.allclose(od["scalar"][0:10], test_data["scalar"])
    assert np.allclose(od["vector"][0:10], test_data["vector"])
    assert np.allclose(od["mat"][0:10], test_data["mat"])
    assert np.allclose(od["scalar"][10:], test_data["scalar"])
    assert np.allclose(od["vector"][10:], test_data["vector"])
    assert np.allclose(od["mat"][10:], test_data["mat"])


def test_getInit():
    test_data = dict(
        scalar=np.random.uniform(size=10),
        vector=np.random.uniform(size=100).reshape(10, 10),
        mat=np.random.uniform(size=1000).reshape(10, 10, 10),
    )

    dd = arrayUtils.getInitializationForODict(test_data)
    assert dd["scalar"][0] == (10,)
    assert dd["vector"][0] == (10, 10)
    assert dd["mat"][0] == (10, 10, 10)
    dd = arrayUtils.getInitializationForODict(test_data, 12)
    assert dd["scalar"][0] == (12,)
    assert dd["vector"][0] == (12, 10)
    assert dd["mat"][0] == (12, 10, 10)


def test_types():
    """Test the typing functions"""
    assert not types.istablelike(4)
    assert types.istablelike(dict(a=np.ones(4)))
    assert not types.istablelike(dict(a=np.ones(4), b=np.ones(5)))
    assert not types.istabledictlike(4)
    assert not types.istabledictlike(dict(a=np.ones(4)))
    assert types.istabledictlike(dict(data=dict(a=np.ones(4))))
    try:
        types.fileType("xx.out")
    except KeyError:
        pass
    else:
        raise KeyError("Failed to catch unknown fileType")


@pytest.mark.skipif(not check_deps([apTable, pd]), reason="Missing an IO package")
def test_type_finders():
    """Test the utils that identify apTables and data frames"""
    import pandas as pd
    from astropy.table import Table

    class DataFrameSub(pd.DataFrame):
        pass

    class TableSub(Table):
        pass

    d1 = pd.DataFrame()
    d2 = DataFrameSub()
    t1 = Table()
    t2 = TableSub()

    assert types.isDataFrame(d1)
    assert types.isDataFrame(d2)
    assert not types.isDataFrame(t1)
    assert not types.isDataFrame(t2)
    assert not types.isDataFrame({})
    assert not types.isDataFrame(77)

    assert not types.isApTable(d1)
    assert not types.isApTable(d2)
    assert types.isApTable(t1)
    assert types.isApTable(t2)
    assert not types.isApTable({})
    assert not types.isApTable(77)


def test_lazy_load():
    """Test that the lazy import works"""
    noModule = lazyImport("thisModuleDoesnotExist")
    try:
        noModule.d
    except ImportError:
        pass
    else:
        raise ImportError("lazyImport failed")


@unittest.skipIf("wave" in sys.modules, "Wave module already imported")
def test_lazy_load2():
    """A second test that the lazy import works"""
    # I picked an obscure python module that is unlikely
    # to be loaded by anything else.

    wave = lazyImport("wave")
    # should not be loaded yet
    assert "wave" not in sys.modules

    # should trigger load
    assert "WAVE_FORMAT_PCM" in dir(wave)
    assert wave.sys == sys

    assert "wave" in sys.modules



"""Type defintions for tables_io"""

import os
from collections import OrderedDict
from collections.abc import Iterable, Mapping

import numpy as np

from .arrayUtils import arrayLength

# Tabular data formats
AP_TABLE = 0
NUMPY_DICT = 1
NUMPY_RECARRAY = 2
PD_DATAFRAME = 3

TABULAR_FORMAT_NAMES = OrderedDict(
    [
        ("astropyTable", AP_TABLE),
        ("numpyDict", NUMPY_DICT),
        ("numpyRecarray", NUMPY_RECARRAY),
        ("pandasDataFrame", PD_DATAFRAME),
    ]
)

TABULAR_FORMATS = OrderedDict([(val, key) for key, val in TABULAR_FORMAT_NAMES.items()])


# File Formats
ASTROPY_FITS = 0
ASTROPY_HDF5 = 1
NUMPY_HDF5 = 2
NUMPY_FITS = 3
PANDAS_HDF5 = 4
PANDAS_PARQUET = 5

FILE_FORMAT_NAMES = OrderedDict(
    [
        ("astropyFits", ASTROPY_FITS),
        ("astropyHdf5", ASTROPY_HDF5),
        ("numpyHdf5", NUMPY_HDF5),
        ("numpyFits", NUMPY_FITS),
        ("pandasHdf5", PANDAS_HDF5),
        ("pandaParquet", PANDAS_PARQUET),
    ]
)

# Default suffixes for various file formats
FILE_FORMAT_SUFFIXS = OrderedDict(
    [
        ("fits", ASTROPY_FITS),
        ("hf5", ASTROPY_HDF5),
        ("hdf5", NUMPY_HDF5),
        ("fit", NUMPY_FITS),
        ("h5", PANDAS_HDF5),
        ("parquet", PANDAS_PARQUET),
        ("parq", PANDAS_PARQUET),
        ("pq", PANDAS_PARQUET),
    ]
)

DEFAULT_TABLE_KEY = OrderedDict(
    [
        ("fits", ""),
        ("hf5", None),
        ("hdf5", None),
        ("fit", ""),
        ("h5", "data"),
        ("parquet", ""),
        ("parq", ""),
        ("pq", ""),
    ]
)

FILE_FORMATS = OrderedDict([(val, key) for key, val in FILE_FORMAT_NAMES.items()])

FILE_FORMAT_SUFFIX_MAP = OrderedDict([(val, key) for key, val in FILE_FORMAT_SUFFIXS.items()])

# Default format to write various table types
NATIVE_FORMAT = OrderedDict(
    [
        (AP_TABLE, ASTROPY_HDF5),
        (NUMPY_DICT, NUMPY_HDF5),
        (NUMPY_RECARRAY, NUMPY_FITS),
        (PD_DATAFRAME, PANDAS_PARQUET),
    ]
)

NATIVE_TABLE_TYPE = OrderedDict([(val, key) for key, val in NATIVE_FORMAT.items()])

# Allowed formats to write various table types
ALLOWED_FORMATS = OrderedDict(
    [
        (AP_TABLE, [ASTROPY_FITS, ASTROPY_HDF5]),
        (NUMPY_DICT, [NUMPY_HDF5]),
        (NUMPY_RECARRAY, [ASTROPY_FITS]),
        (PD_DATAFRAME, [PANDAS_PARQUET, PANDAS_HDF5]),
    ]
)


def isDataFrame(obj):
    for c in obj.__class__.__mro__:
        if c.__name__ == "DataFrame" and c.__module__ == "pandas.core.frame":
            return True
    return False


def isApTable(obj):
    for c in obj.__class__.__mro__:
        if c.__name__ == "Table" and c.__module__ == "astropy.table.table":
            return True
    return False


def tableType(obj):
    """Identify the type of table we have

    Parameters
    ----------
    obj : `object`
        The input object

    Returns
    -------
    otype : `int`
        The object type, one of `TABULAR_FORMATS.keys()`

    Raises
    ------
    TypeError
        The object is not a supported type
    IndexError
        One of the columns in a Mapping is the wrong length
    """
    if isDataFrame(obj):
        return PD_DATAFRAME
    if isApTable(obj):
        return AP_TABLE
    if isinstance(obj, np.recarray):
        return NUMPY_RECARRAY
    if not isinstance(obj, Mapping):
        raise TypeError(
            f"Object of type {type(obj)} is not one of the supported types"
            f"{list(TABULAR_FORMAT_NAMES.keys())}"
        )

    nRow = None
    for key, val in obj.items():
        if istablelike(val):
            raise TypeError(f"Column {key} is a table of type {type(val)}")
        if not isinstance(val, Iterable):  # pragma: no cover
            raise TypeError(f"Column {key} of type {type(val)} is not iterable")
        if nRow is None:
            nRow = arrayLength(val)
        else:
            if arrayLength(val) != nRow:
                raise IndexError(
                    f"Column {key} length {arrayLength(val)} != {nRow}"
                )  # pylint: disable=bad-string-format-type
    return NUMPY_DICT


def istablelike(obj):
    """Test to see if an object is one of the supported table types

    Parameters
    ----------
    obj : `object`
        The input object

    Returns
    -------
    tablelike : `bool`
        True is the object is `Tablelike`, False otherwise
    """
    try:
        _ = tableType(obj)
    except (TypeError, IndexError):
        return False
    return True


def istabledictlike(obj):
    """Test to see if an object is a `Mapping`, (`str`, `Tablelike`)

    Parameters
    ----------
    obj : `object`
        The input object

    Returns
    -------
    tabledict : `bool`
        True is the object is a `Mapping`, (`str`, `Tablelike`), False otherwise
    """
    if not isinstance(obj, Mapping):
        return False
    for val in obj.values():
        if not istablelike(val):
            return False
    return True


def fileType(filepath, fmt=None):
    """Identify the type of file we have

    Parameters
    ----------
    filepath : `str`
        The path to the file

    fmt : `str` or `None`
        Overrides the file extension

    Returns
    -------
    otype : `int`
        The object type, one of `FILE_FORMATS.keys()`

    Raises
    ------
    KeyError
        The file format is not a support value
    """
    if fmt is None:
        fmt = os.path.splitext(filepath)[1][1:]
    try:
        return FILE_FORMAT_SUFFIXS[fmt]
    except KeyError as msg:
        raise KeyError(
            f"Unknown file format {fmt}, supported types are" f"{list(FILE_FORMAT_SUFFIXS.keys())}"
        ) from msg



"""
Unit tests for io_layer module
"""

import pytest

from tables_io import types, convert, TableDict
from tables_io.testUtils import compare_table_dicts, check_deps
from tables_io.lazy_modules import apTable, jnp, h5py, pq


@pytest.mark.skipif(not check_deps([apTable, h5py, pq, jnp]), reason="Missing an IO package")
def test_bad_type(data_tables):
    """Test adding bad types to TableDict"""
    td = TableDict(data_tables)
    with pytest.raises(TypeError, match="item 4 was not recognized as a table."):
        td["aa"] = 4


@pytest.mark.skipif(not check_deps([apTable, h5py, pq, jnp]), reason="Missing an IO package")
@pytest.mark.parametrize(
    "tType, fmt, use_keys",
    [
        (types.AP_TABLE, "fits", False),
        (types.AP_TABLE, "hf5", False),
        (types.NUMPY_DICT, "hdf5", False),
        (types.PD_DATAFRAME, "h5", False),
        (types.PD_DATAFRAME, "pq", True),
    ],
)
def test_parquet_dict(data_tables, data_table, data_keys, tmp_path, tType, fmt, use_keys):
    """Test writing / reading {tType} dataframes to {fmt}"""
    basepath = str(tmp_path)

    odict_c = convert(data_tables, tType)
    td_c = TableDict(odict_c)
    filepath = td_c.write(basepath, fmt)
    td_r = TableDict.read(filepath, tType=tType, fmt=fmt, keys=data_keys if use_keys else None)
    tables_r = td_r.convert(types.AP_TABLE)
    assert compare_table_dicts(data_tables, tables_r)
    if fmt in ["pq", "h5"]:
        return
    basepath2 = "%s_v2" % basepath
    filepath2 = td_c.write(basepath2)
    td_r2 = TableDict.read(filepath2, tType=tType)
    tables_r2 = td_r2.convert(types.AP_TABLE)
    assert compare_table_dicts(data_tables, tables_r2)



"""Array-related utility functions for tables_io"""

from collections import OrderedDict

import numpy as np


def arrayLength(arr):
    """Get the length of an array

    The works on scalars and arrays, so it is safe to use

    For scalars it returns 0
    For arrays it np.shape(arr)[0]

    Parameters
    ----------
    arr : `array-like`
        The input array

    Returns
    -------
    length : `int`
        The object length
    """
    shape = np.shape(arr)
    if not shape:
        return 0
    return shape[0]


def forceToPandables(arr, check_nrow=None):
    """
    Forces a  `numpy.array` into a format that pandas can handle

    Parameters
    ----------
    arr : `numpy.array`
        The input array

    check_nrow : `int` or `None`
        If not None, require that `arr.shape[0]` match this value

    Returns
    -------
    out : `numpy.array` or `list` of `numpy.array`
        Something that pandas can handle
    """
    ndim = np.ndim(arr)
    if ndim == 0:
        return arr
    shape = np.shape(arr)
    nrow = shape[0]
    if check_nrow is not None and check_nrow != nrow:
        raise ValueError(f"Number of rows does not match: {nrow} != {check_nrow}")
    if ndim == 1:
        return arr
    if ndim == 2:
        return list(arr)
    shape = np.shape(arr)
    ncol = np.prod(shape[1:])
    return list(arr.reshape(nrow, ncol))


def getGroupInputDataLength(hg):
    """Return the length of a HDF5 group

    Parameters
    ----------
    hg : `h5py.Group` or `h5py.File`
        The input data group

    Returns
    -------
    length : `int`
        The length of the data

    Notes
    -----
    For a multi-D array this return the length of the first axis
    and not the total size of the array.

    Normally that is what you want to be iterating over.

    The group is meant to represent a table, hence all child datasets
    should be the same length
    """
    firstkey = list(hg.keys())[0]
    nrows = len(hg[firstkey])
    firstname = hg[firstkey].name
    for value in hg.values():
        if len(value) != nrows:
            raise ValueError(
                f"Group does not represent a table. Length ({len(value)})"
                f"of column {value.name} not not match length ({nrows}) of"
                f"first column {firstname}"
            )
    return nrows


def getInitializationForODict(in_dict, nRow_out=None):
    """shape of arrays and dtypes in a dictionary.
    This is useful for initialize hdf5 files

    Parameters
    ----------
    in_dict : `dict`
        The dictionary with the data
    nRow_out : `int`
        The number of rows in the output

    Returns
    -------
    out_dict : dict (str, tuple(tuple, dtype))
    """
    out_dict = {}
    for key, val in in_dict.items():
        shape = list(np.shape(val))
        if nRow_out is not None:
            shape[0] = nRow_out
        out_dict[key] = (tuple(shape), val.dtype)
    return out_dict


def printDictShape(in_dict):
    """Print the shape of arrays in a dictionary.
    This is useful for debugging `astropy.Table` creation.

    Parameters
    ----------
    in_dict : `dict`
        The dictionary to print
    """
    for key, val in in_dict.items():
        print(key, np.shape(val))


def sliceDict(in_dict, subslice):
    """Create a new `dict` by taking a slice of of every array in a `dict`

    Parameters
    ----------
    in_dict : `dict`
        The dictionary to extract from
    subslice : `int` or `slice`
        Used to slice the arrays

    Returns
    -------
    out_dict : `dict`
        The converted dicionary
    """

    out_dict = OrderedDict()
    for key, val in in_dict.items():
        try:
            out_dict[key] = val[subslice]
        except (KeyError, TypeError):  # pragma: no cover
            out_dict[key] = val
    return out_dict


def checkKeys(in_dicts):
    """Check that the keys in all the in_dicts match

    Parameters
    ----------
    in_dicts : `list`, (`OrderedDict`, (`str`, `numpy.array`))
        The dictionaries for which compare keys

    Raises
    ------
    KeyError
        if keys do not match
    """
    if not in_dicts:  # pragma: no cover
        return
    master_keys = in_dicts[0].keys()
    for in_dict in in_dicts[1:]:
        if in_dict.keys() != master_keys:  # pragma: no cover
            raise ValueError(f"Keys do not match: {list(in_dict.keys())} != {list(master_keys)}")


def concatenateDicts(in_dicts):
    """Create a new `dict` by concatenating each array in `in_dicts`

    Parameters
    ----------
    in_dicts : `list`, (`OrderedDict`, (`str`, `numpy.array`))
        The dictionaries to stack

    Returns
    -------
    out_dict : `dict`
        The stacked dicionary
    """
    if not in_dicts:  # pragma: no cover
        return OrderedDict()
    checkKeys(in_dicts)
    out_dict = OrderedDict([(key, None) for key in in_dicts[0].keys()])
    for key in out_dict.keys():
        out_dict[key] = np.concatenate([in_dict[key] for in_dict in in_dicts])
    return out_dict



"""IO Functions for tables_io"""

from collections import OrderedDict

import numpy as np

from .arrayUtils import forceToPandables
from .lazy_modules import apTable, fits, pd
from .types import AP_TABLE, NUMPY_DICT, NUMPY_RECARRAY, PD_DATAFRAME, istablelike, tableType

### I. Single `Tablelike` conversions

### I A. Converting to `astropy.table.Table`


def dataFrameToApTable(df):
    """
    Convert a `pandas.DataFrame` to an `astropy.table.Table`

    Parameters
    ----------
    df :  `pandas.DataFrame`
        The dataframe

    Returns
    -------
    tab : `astropy.table.Table`
        The table
    """
    o_dict = OrderedDict()
    for colname in df.columns:
        col = df[colname]
        if col.dtype.name == "object":
            o_dict[colname] = np.vstack(col.to_numpy())
        else:
            o_dict[colname] = col.to_numpy()
    tab = apTable.Table(o_dict)
    for k, v in df.attrs.items():
        tab.meta[k] = v  # pragma: no cover
    return tab


def convertToApTable(obj):
    """
    Convert an object to an `astropy.table.Table`

    Parameters
    ----------
    obj : `object`
       The object being converted

    Returns
    -------
    tab : `astropy.table.Table`
        The table
    """
    tType = tableType(obj)
    if tType == AP_TABLE:
        return obj
    if tType == NUMPY_DICT:
        return apTable.Table(obj)
    if tType == NUMPY_RECARRAY:
        return apTable.Table(obj)
    if tType == PD_DATAFRAME:
        # try this: apTable.from_pandas(obj)
        return dataFrameToApTable(obj)
    raise TypeError(f"Unsupported TableType {tType}")  # pragma: no cover


### I B. Converting to `OrderedDict`, (`str`, `numpy.array`)


def apTableToDict(tab):
    """
    Convert an `astropy.table.Table` to an `OrderedDict` of `str` : `numpy.array`

    Parameters
    ----------
    tab :  `astropy.table.Table`
        The table

    Returns
    --------
    data : `OrderedDict`,  (`str` : `numpy.array`)
        The tabledata
    """
    data = OrderedDict()
    for key, val in zip(tab.colnames, tab.itercols()):
        data[key] = np.array(val)
    return data


def recarrayToDict(rec):
    """
    Convert an `np.recarray` to an `OrderedDict` of `str` : `numpy.array`

    Parameters
    ----------
    rec :  `np.recarray`
        The input recarray

    Returns
    --------
    data : `OrderedDict`,  (`str` : `numpy.array`)
        The tabledata
    """
    return OrderedDict([(colName, rec[colName]) for colName in rec.dtype.names])


def dataFrameToDict(df):
    """
    Convert a `pandas.DataFrame` to an `OrderedDict` of `str` : `numpy.array`

    Parameters
    ----------
    df :  `pandas.DataFrame`
        The dataframe

    Returns
    --------
    data : `OrderedDict`,  (`str` : `numpy.array`)
        The tabledata
    """
    data = OrderedDict()
    for key in df.keys():
        col = df[key]
        if col.dtype.name == "object":
            data[key] = np.vstack(col.to_numpy())
        else:
            data[key] = np.array(col)
    return data


def hdf5GroupToDict(hg):
    """
    Convert a `hdf5` object to an `OrderedDict`, (`str`, `numpy.array`)

    Parameters
    ----------
    hg :  `h5py.File` or `h5py.Group`
        The hdf5 object

    Returns
    --------
    data : `OrderedDict`,  (`str` : `numpy.array`)
        The tabledata
    """
    data = OrderedDict()
    for key in hg.keys():
        data[key] = np.array(hg[key])
    return data


def convertToDict(obj):
    """
    Convert an object to an `OrderedDict`, (`str`, `numpy.array`)

    Parameters
    ----------
    obj : `object`
       The object being converted

    Returns
    -------
    tab : `astropy.table.Table`
        The table
    """
    tType = tableType(obj)
    if tType == AP_TABLE:
        return apTableToDict(obj)
    if tType == NUMPY_DICT:
        return obj
    if tType == NUMPY_RECARRAY:
        return recarrayToDict(obj)
    if tType == PD_DATAFRAME:
        return dataFrameToDict(obj)
    raise TypeError(f"Unsupported TableType {tType}")  # pragma: no cover


### I C. Converting to `np.recarray`


def apTableToRecarray(tab):
    """
    Convert an `astropy.table.Table` to an `numpy.recarray`

    Parameters
    ----------
    tab :  `astropy.table.Table`
        The table

    Returns
    --------
    rec : `numpy.recarray`
        The output rec array
    """
    return fits.table_to_hdu(tab).data


def convertToRecarray(obj):
    """
    Convert an object to an `numpy.recarray`

    Parameters
    ----------
    obj : `object`
       The object being converted

    Returns
    -------
    rec : `numpy.recarray`
        The output recarray
    """
    tType = tableType(obj)
    if tType == AP_TABLE:
        return apTableToRecarray(obj)
    if tType == NUMPY_DICT:
        return apTableToRecarray(apTable.Table(obj))
    if tType == NUMPY_RECARRAY:
        return obj
    if tType == PD_DATAFRAME:
        return apTableToRecarray(dataFrameToApTable(obj))
    raise TypeError(f"Unsupported TableType {tType}")  # pragma: no cover


### I D. Converting to `pandas.DataFrame`


def apTableToDataFrame(tab):
    """
    Convert an `astropy.table.Table` to a `pandas.DataFrame`

    Parameters
    ----------
    tab : `astropy.table.Table`
        The table

    Returns
    -------
    df :  `pandas.DataFrame`
        The dataframe
    """
    o_dict = OrderedDict()
    for colname in tab.columns:
        col = tab[colname]
        o_dict[colname] = forceToPandables(col.data)
    df = pd.DataFrame(o_dict)
    for k, v in tab.meta.items():
        df.attrs[k] = v  # pragma: no cover
    return df


def dictToDataFrame(odict, meta=None):
    """
    Convert an `OrderedDict`, (`str`, `numpy.array`) to a `pandas.DataFrame`

    Parameters
    ----------
    odict : `OrderedDict`, (`str`, `numpy.array`)
        The dict

    meta : `dict` or `None`
        Optional dictionary of metadata

    Returns
    -------
    df :  `pandas.DataFrame`
        The dataframe
    """
    outdict = OrderedDict()
    for k, v in odict.items():
        outdict[k] = forceToPandables(v)
    df = pd.DataFrame(outdict)
    if meta is not None:  # pragma: no cover
        for k, v in meta.items():
            df.attrs[k] = v
    return df


def convertToDataFrame(obj):
    """
    Convert an object to a `pandas.DataFrame`

    Parameters
    ----------
    obj : `object`
       The object being converted

    Returns
    -------
    df :  `pandas.DataFrame`
        The dataframe
    """
    tType = tableType(obj)
    if tType == AP_TABLE:
        return apTableToDataFrame(obj)
    if tType == NUMPY_DICT:
        return dictToDataFrame(obj)
    if tType == NUMPY_RECARRAY:
        odict = recarrayToDict(obj)
        return dictToDataFrame(odict)
    if tType == PD_DATAFRAME:
        return obj
    raise TypeError(f"Unsupported tableType {tType}")  # pragma: no cover


# I E. Generic `convert`


def convertObj(obj, tType):
    """
    Convert an object to a specific type of `Tablelike`

    Parameters
    ----------
    obj : `object`
       The object being converted
    tType : `int`
       The type of object to convert to, one of `TABULAR_FORMAT_NAMES.keys()`

    Returns
    -------
    out :  `Tablelike`
        The converted object
    """
    if tType == AP_TABLE:
        return convertToApTable(obj)
    if tType == NUMPY_DICT:
        return convertToDict(obj)
    if tType == NUMPY_RECARRAY:
        return convertToRecarray(obj)
    if tType == PD_DATAFRAME:
        return convertToDataFrame(obj)
    raise TypeError(f"Unsupported tableType {tType}")  # pragma: no cover


### II.  Multi-table conversion utilities


def convertToApTables(odict):
    """
    Convert several `objects` to `astropy.table.Table`

    Parameters
    ----------
    odict :  `Mapping`, (`str`, `Tablelike`)
        The input objects

    Returns
    -------
    tabs : `OrderedDict` of `astropy.table.Table`
        The tables
    """
    return OrderedDict([(k, convertToApTable(v)) for k, v in odict.items()])


def convertToDicts(odict):
    """
    Convert several `objects` to `OrderedDict`, (`str`, `numpy.array`)

    Parameters
    ----------
    odict :  `Mapping`, (`str`, `Tablelike`)
        The input objects

    Returns
    -------
    tabs : `OrderedDict` of `OrderedDict`, (`str`, `numpy.array`)
        The tables
    """
    return OrderedDict([(k, convertToDict(v)) for k, v in odict.items()])


def convertToRecarrays(odict):
    """
    Convert several `objects` to `np.recarray`

    Parameters
    ----------
    odict :  `Mapping`, (`str`, `Tablelike`)
        The input objects

    Returns
    -------
    tabs : `OrderedDict` of `np.recarray`
        The tables
    """
    return OrderedDict([(k, convertToRecarray(v)) for k, v in odict.items()])


def convertToDataFrames(odict):
    """
    Convert several `objects` to `pandas.DataFrame`

    Parameters
    ----------
    odict :  `Mapping`, (`str`, `Tablelike`)
        The input objects

    Returns
    -------
    df :  `OrderedDict` of `pandas.DataFrame`
        The dataframes
    """
    return OrderedDict([(k, convertToDataFrame(v)) for k, v in odict.items()])


def convert(obj, tType):
    """
    Convert several `objects` to a specific type

    Parameters
    ----------
    obj :  'Tablelike` or `TableDictlike`
        The input object

    tType : `int`
        One of `TABULAR_FORMAT_NAMES.keys()`

    Returns
    -------
    out :  `Tablelike` or `TableDictlike`
        The converted data
    """
    if istablelike(obj):
        return convertObj(obj, tType)

    funcMap = {
        AP_TABLE: convertToApTables,
        NUMPY_DICT: convertToDicts,
        NUMPY_RECARRAY: convertToRecarrays,
        PD_DATAFRAME: convertToDataFrames,
    }
    try:
        theFunc = funcMap[tType]
    except KeyError as msg:  # pragma: no cover
        raise KeyError(f"Unsupported type {tType}") from msg
    return theFunc(obj)



"""
Unit tests for io_layer module
"""

import pytest
from tables_io import types, convert, convertObj
from tables_io.testUtils import compare_table_dicts, check_deps
from tables_io.lazy_modules import tables, apTable, apDiffUtils, fits, h5py, pd, pq, jnp


@pytest.mark.parametrize(
    "mod",
    [tables, apTable, apDiffUtils, fits, h5py, pd, pq],
)
def test_deps(mod):
    assert check_deps([mod])


@pytest.mark.skipif(not check_deps([jnp]), reason="Failed to load jax.numpy")
def test_deps_jnp():
    assert check_deps([jnp])


def test_bad_deps():
    dummy = 0
    assert not check_deps([dummy])
    

@pytest.mark.skipif(not check_deps([apTable, pd]), reason="Missing panda or astropy.table")
@pytest.mark.parametrize(
    "tType1",
    [types.AP_TABLE, types.NUMPY_DICT, types.NUMPY_RECARRAY, types.PD_DATAFRAME],
)
@pytest.mark.parametrize(
    "tType2",
    [types.AP_TABLE, types.NUMPY_DICT, types.NUMPY_RECARRAY, types.PD_DATAFRAME],
)
def test_type_conversion(data_tables, data_table, tType1, tType2):
    """Perform type conversion on the cross-product of all types."""
    odict_1 = convert(data_tables, tType1)
    odict_2 = convert(odict_1, tType2)
    tables_r = convert(odict_2, types.AP_TABLE)
    assert compare_table_dicts(data_tables, tables_r)
    t1 = convertObj(data_table, tType1)
    t2 = convertObj(t1, tType2)
    _ = convertObj(t2, types.AP_TABLE)
